import RPi.GPIO as GPIO
import time
import enum
from typing import Any, Optional, Dict

class GPIORelaisModel():

    defaultValue = GPIO.HIGH

    def __init__(
        self, uid: int,
        pin: Optional[int],
        name: Optional[str] = None,
        description: Optional[str] = None
    ):
        self.uid = uid
        self.pin = pin
        self.name = name
        self.description = description

    def setDefaultValue(self, value: int):
        self.defaultValue = value

    def toDefault(self):
        GPIO.output(self.pin, self.defaultValue)

    def getStatus(self):
        return GPIO.input(self.pin)

    def setStatus(self, value):
        GPIO.output(self.pin, value)
        return self.getStatus()

    def to_dict(self):
        return {
            "uid": self.uid,
            "pin": self.pin,
            "defaultValue": self.defaultValue,
            "status": self.getStatus(),
            "name": self.name,
            "description": self.description
        }


class GPIORelaisAdapter():
    @staticmethod
    def getGPIORelaisFor(data) -> GPIORelaisModel:
        name = None
        description = None
        if "params" in data:            
            params = data["params"]
            if "name" in params:
                name = params["name"]
            if "description" in params:
                description = params["description"]

        model = GPIORelaisModel(0, int(data["id"]), name, description)
        model.setDefaultValue(int(data["defaultValue"]))
        return model


class GPIOStoppingPoint(GPIORelaisModel):

    def __init__(
        self, uid: int,
        pin: Optional[int], measurmentpin: Optional[int],
        name: Optional[str], description: Optional[str]
    ):
        self.measurmentpin = measurmentpin
        super(GPIOStoppingPoint, self).__init__(uid, pin, name, description)

    def to_dict(self):
        mdict = super(GPIOStoppingPoint, self).to_dict()
        mdict["measurmentpin"] = self.measurmentpin
        mdict["uid"] = self.uid
        return mdict

    @classmethod
    def fromParent(cls, parent: GPIORelaisModel, measurmentpin: Optional[int]):
        classInstance = cls(parent.uid, parent.pin, measurmentpin, parent.name, parent.description)
        classInstance.setDefaultValue(parent.defaultValue)
        return classInstance

    @classmethod
    def from_dict(cls, data: Dict[str, Any], id: int):
        stop = cls(id, data.get("pin"), data.get("measurmentpin"), data.get("name"), data.get("description"))
        stop.defaultValue = data.get("defaultValue")
        return stop


class GPIOSwitchType(enum.Enum):
    TM_SWITCH_STRAIGHT_LEFT = 1
    TM_SWITCH_STRAIGHT_RIGHT = 2
    TM_SWITCH_BORN_LEFT = 3
    TM_SWITCH_BORN_RIGHT = 4
    TM_SWITCH_CROSS = 5


class GPIOSwitchPoint(GPIORelaisModel):

    def __init__(
        self, uid: int,
        switchType: Optional[str], pin: Optional[int],
        name: Optional[str], description: Optional[str]
    ):
        self.needsPowerOn = True
        self.switchType = switchType
        self.powerRelais = None
        super(GPIOSwitchPoint, self).__init__(uid, pin, name, description)

    def setPowerRelais(self, relais: GPIORelaisModel):
        self.powerRelais = relais

    def setStatus(self, value: int):
        if self.needsPowerOn and self.powerRelais is not None:
            self.powerRelais.setStatus(GPIO.LOW)
        GPIO.output(self.pin, value)
        time.sleep(0.2)
        if self.needsPowerOn and self.powerRelais is not None:
            self.powerRelais.setStatus(GPIO.HIGH)
        return self.getStatus()

    def toDefault(self):
        if self.needsPowerOn and self.powerRelais is not None:
            self.powerRelais.setStatus(GPIO.LOW)
        GPIO.output(self.pin, self.defaultValue)
        time.sleep(0.2)
        if self.needsPowerOn and self.powerRelais is not None:
            self.powerRelais.setStatus(GPIO.HIGH)

    def to_dict(self):
        mdict = super(GPIOSwitchPoint, self).to_dict()
        mdict["needsPowerOn"] = self.needsPowerOn
        mdict["switchType"] = self.switchType
        mdict["uid"] = self.uid
        return mdict

    @classmethod
    def from_dict(cls, data: Dict[str, Any], id: int):
        params = data.get("params")
        switchType = None
        if params is not None:
            switchType = params.get("switchType")            
            if GPIOSwitchHelper.isValidType(switchType) is False:
                raise ValueError("Invalid switch type")
        switch = cls(id, switchType, data.get("pin"), data.get("name"), data.get("description"))
        switch.defaultValue = data.get("defaultValue")
        switch.needsPowerOn = data.get("needsPowerOn")
        return switch

    @classmethod
    def fromParent(cls, parent: GPIORelaisModel, switchType: str):
        classInstance = cls(parent.uid, switchType, parent.pin, parent.name, parent.description)
        classInstance.setDefaultValue(parent.defaultValue)
        return classInstance


class GPIOSwitchHelper():

    @staticmethod
    def isValidType(type: str) -> bool:
        for mType in (GPIOSwitchType):
            if mType.name == type:
                return True
        return False



from flask import Blueprint
from flask import request
from flask import abort
from . import baseAPI
from .validator import Validator
from . import apiController
import json
from .databaseControllerModule import DatabaseController

configApi = Blueprint('configApi', __name__)

##
# Endpoints Config
##

@configApi.route('/trainmote/api/v1/config', methods=["GET"])
def getConfig():
    config = DatabaseController().getConfig()
    if config is not None:
        return json.dumps(config.to_dict()), 200, baseAPI.defaultHeader()
    else:
        abort(404)

@configApi.route('/trainmote/api/v1/config', methods=["POST"])
def setConfig():
    mJson = request.get_json()
    if mJson is not None:
        validator = Validator()
        if validator.validateDict(mJson, "config_scheme") is False:
            abort(400)

        stops = DatabaseController().getAllStopModels()
        switchs = DatabaseController().getAllSwichtModels()
        switchPowerRelaisIsStop = validator.containsPin(int(mJson["switchPowerRelais"]), stops)
        switchPowerRelaisIsSwitch = validator.containsPin(int(mJson["switchPowerRelais"]), switchs)
        if switchPowerRelaisIsStop or switchPowerRelaisIsSwitch:
            return json.dumps({"error": "Switch power relais pin is already in use"}), 409, baseAPI.defaultHeader()

        powerRelaisIsStop = validator.containsPin(int(mJson["powerRelais"]), stops)
        powerRelaisIsSwitch = validator.containsPin(int(mJson["powerRelais"]), switchs)
        if powerRelaisIsStop or powerRelaisIsSwitch:
            return json.dumps({"error": "Power relais pin is already in use"}), 409, baseAPI.defaultHeader()

        stateRelaisIsStop = validator.containsPin(int(mJson["stateRelais"]), stops)
        stateRelaisIsSwitch = validator.containsPin(int(mJson["stateRelais"]), switchs)
        if stateRelaisIsStop or stateRelaisIsSwitch:
            return json.dumps({"error": "State relais pin is already in use"}), 409, baseAPI.defaultHeader()

        DatabaseController().insertConfig(
            int(mJson["switchPowerRelais"]),
            int(mJson["powerRelais"]),
            int(mJson["stateRelais"])
        )

        if apiController.powerThread is not None:
            apiController.powerThread.stop()

        config = DatabaseController().getConfig()
        if config.powerRelais is not None:
            apiController.setupPowerGPIO(config.powerRelais)
        if config is not None:
            return json.dumps(config.to_dict()), 201, baseAPI.defaultHeader()
        else:
            abort(500)
    else:
        abort(400)



import board
import busio
import adafruit_ads1x15.ads1115 as ADS
from adafruit_ads1x15.analog_in import AnalogIn
import time
import threading


class TrackingService:

    isTracking = False

    def __init__(self, stoppingPoint):
        self.stoppingPoint = stoppingPoint
        self.trackingThread = TrackerThread(stoppingPoint)

    def startTracking(self):
        # do some stuff
        self.isTracking = True
        self.trackingThread.start()
        print('Start Tracking: ', self.stoppingPoint.measurmentpin)
        # continue doing stuff

    def stopTracking(self):
        print('Stop Tracking: ', self.stoppingPoint.measurmentpin)
        self.trackingThread.kill.set()
        self.trackingThread.join()
        self.isTracking = False


class TrackerThread(threading.Thread):

    GAIN = 1
    # Create the I2C bus
    i2c = busio.I2C(board.SCL, board.SDA)

    def __init__(self, stoppingPoint):
        threading.Thread.__init__(self)
        self.stoppingPoint = stoppingPoint
        self.kill = threading.Event()

        try:
            # Create the ADC object using the I2C bus
            ads = ADS.ADS1115(self.i2c)
        except ValueError:
            print("Oops! ADS1115 not installed")

    def run(self):
        self.trackVoltage()

    def trackVoltage(self):
        while not self.kill.is_set():
            # chan = AnalogIn(ads, ADS.P0)
            # currentVoltage = self.ads.read_adc(self.stoppingPoint.measurmentpin, gain= self.GAIN)
            # if abs(currentVoltage) > 10:
            # print ('Detected voltage at Stopping Point: ', self.stoppingPoint.measurmentpin)
            time.sleep(0.3)



from sqlite3.dbapi2 import Error
from flask import Blueprint
from flask import request
from flask import abort
from flask import Response

from pkg_trainmote.models.GPIORelaisModel import GPIOSwitchPoint
from . import baseAPI
from . import gpioservice
from .validator import Validator
import json
from .databaseControllerModule import DatabaseController

switchApiBlueprint = Blueprint('switchApi', __name__)

##
# Endpoint Switch
##

@switchApiBlueprint.route('/trainmote/api/v1/switch/<switch_id>', methods=["GET"])
def switch(switch_id: str):
    if switch_id is None:
        abort(400)
    try:
        return gpioservice.getSwitch(switch_id), 200, baseAPI.defaultHeader()
    except ValueError as e:
        return json.dumps({"error": str(e)}), 404, baseAPI.defaultHeader()


@switchApiBlueprint.route('/trainmote/api/v1/control/switch/<switch_id>', methods=["PATCH"])
def setSwitch(switch_id: str):
    if switch_id is None:
        abort(400)
    try:
        return gpioservice.setSwitch(switch_id), 200, baseAPI.defaultHeader()
    except ValueError as e:
        return json.dumps({"error": str(e)}), 400, baseAPI.defaultHeader()


@switchApiBlueprint.route('/trainmote/api/v1/switch/<switch_id>', methods=["PATCH"])
def updateSwitch(switch_id: str):
    mJson = request.get_json()
    if mJson is not None:
        validator = Validator()
        if validator.validateDict(mJson, "switch_update_scheme") is False:
            abort(400)
        try:
            database = DatabaseController()
            exModel = database.getSwitch(int(switch_id))
            if exModel is None:
                return json.dumps({"error": "Switch for id {} not found".format(switch_id)}), 404
            model = GPIOSwitchPoint.from_dict(mJson, int(switch_id))
            if model.pin is not None and exModel.pin is not None and model.pin is not exModel.pin:
                validator.isAlreadyInUse(int(mJson["pin"]))
            updatedSwitch = database.updateSwitch(int(switch_id), model)
            if updateSwitch is not None:
                return json.dumps({"model": updatedSwitch.to_dict()})
            else:
                abort(500)

        except ValueError as e:
            return json.dumps({"error": str(e)}), 409, baseAPI.defaultHeader()
        except Error as e:
            return json.dumps({"error": str(e)}), 400, baseAPI.defaultHeader()
    else:
        abort(400)


@switchApiBlueprint.route('/trainmote/api/v1/switch/<switch_id>', methods=["DELETE"])
def deleteSwitch(switch_id: str):
    if switch_id is None:
        abort(400)
    DatabaseController().deleteSwitchModel(int(switch_id)), 205, baseAPI.defaultHeader()
    return 'ok'


@switchApiBlueprint.route('/trainmote/api/v1/switch', methods=["POST"])
def addSwitch():
    mJson = request.get_json()
    if mJson is not None:
        if Validator().validateDict(mJson, "switch_scheme") is False:
            abort(400)
        config = DatabaseController().getConfig()
        if config is not None and config.containsPin(mJson["id"]):
            return json.dumps({"error": "Pin is already in use as power relais"}), 409, baseAPI.defaultHeader()

        try:
            return gpioservice.createSwitch(mJson), 201, baseAPI.defaultHeader()
        except ValueError as e:
            return json.dumps({"error": str(e)}), 400, baseAPI.defaultHeader()
    else:
        abort(400)


@switchApiBlueprint.route('/trainmote/api/v1/switch/all')
def getAllSwitches():
    return Response(gpioservice.getAllSwitches(), mimetype="application/json"), 200, baseAPI.defaultHeader()
